require_relative 'helper'
require 'roundhouse/cli'
require 'tempfile'

class Roundhouse::CLI
  def die(code)
    @code = code
  end

  def valid?
    !@code
  end
end

class TestCli < Roundhouse::Test
  describe 'CLI#parse' do

    before do
      @cli = Roundhouse::CLI.new
      @opts = Roundhouse.options.dup
    end

    after do
      Roundhouse.options = @opts
    end

    it 'does not require the specified Ruby code' do
      @cli.parse(['roundhouse', '-r', './test/fake_env.rb'])
      refute($LOADED_FEATURES.any? { |x| x =~ /fake_env/ })
      assert @cli.valid?
    end

    it 'does not boot rails' do
      refute defined?(::Rails::Application)
      @cli.parse(['roundhouse', '-r', './myapp'])
      refute defined?(::Rails::Application)
    end

    it 'changes concurrency' do
      @cli.parse(['roundhouse', '-c', '60', '-r', './test/fake_env.rb'])
      assert_equal 60, Roundhouse.options[:concurrency]
    end

    it 'accepts a process index' do
      @cli.parse(['roundhouse', '-i', '7', '-r', './test/fake_env.rb'])
      assert_equal 7, Roundhouse.options[:index]
    end

    it 'accepts a stringy process index' do
      @cli.parse(['roundhouse', '-i', 'worker.7', '-r', './test/fake_env.rb'])
      assert_equal 7, Roundhouse.options[:index]
    end

    it 'changes timeout' do
      @cli.parse(['roundhouse', '-t', '30', '-r', './test/fake_env.rb'])
      assert_equal 30, Roundhouse.options[:timeout]
    end

    it 'sets verbose' do
      old = Roundhouse.logger.level
      @cli.parse(['roundhouse', '-v', '-r', './test/fake_env.rb'])
      assert_equal Logger::DEBUG, Roundhouse.logger.level
      # If we leave the logger at DEBUG it'll add a lot of noise to the test output
      Roundhouse.options.delete(:verbose)
      Roundhouse.logger.level = old
    end

    describe 'with logfile' do
      before do
        @old_logger = Roundhouse.logger
        @tmp_log_path = '/tmp/roundhouse.log'
      end

      after do
        Roundhouse.logger = @old_logger
        Roundhouse.options.delete(:logfile)
        File.unlink @tmp_log_path if File.exist?(@tmp_log_path)
      end

      it 'sets the logfile path' do
        @cli.parse(['roundhouse', '-L', @tmp_log_path, '-r', './test/fake_env.rb'])

        assert_equal @tmp_log_path, Roundhouse.options[:logfile]
      end

      it 'creates and writes to a logfile' do
        @cli.parse(['roundhouse', '-L', @tmp_log_path, '-r', './test/fake_env.rb'])

        Roundhouse.logger.info('test message')

        assert_match(/test message/, File.read(@tmp_log_path), "didn't include the log message")
      end

      it 'appends messages to a logfile' do
        File.open(@tmp_log_path, 'w') do |f|
          f.puts 'already existent log message'
        end

        @cli.parse(['roundhouse', '-L', @tmp_log_path, '-r', './test/fake_env.rb'])

        Roundhouse.logger.info('test message')

        log_file_content = File.read(@tmp_log_path)
        assert_match(/already existent/, log_file_content, "didn't include the old message")
        assert_match(/test message/, log_file_content, "didn't include the new message")
      end
    end

    describe 'with pidfile' do
      before do
        @tmp_file = Tempfile.new('roundhouse-test')
        @tmp_path = @tmp_file.path
        @tmp_file.close!

        @cli.parse(['roundhouse', '-P', @tmp_path, '-r', './test/fake_env.rb'])
      end

      after do
        File.unlink @tmp_path if File.exist? @tmp_path
      end

      it 'sets pidfile path' do
        assert_equal @tmp_path, Roundhouse.options[:pidfile]
      end

      it 'writes pidfile' do
        assert_equal File.read(@tmp_path).strip.to_i, Process.pid
      end
    end

    describe 'with config file' do
      before do
        @cli.parse(['roundhouse', '-C', './test/config.yml'])
      end

      it 'takes a path' do
        assert_equal './test/config.yml', Roundhouse.options[:config_file]
      end

      it 'sets verbose' do
        refute Roundhouse.options[:verbose]
      end

      it 'sets require file' do
        assert_equal './test/fake_env.rb', Roundhouse.options[:require]
      end

      it 'does not set environment' do
        assert_equal nil, Roundhouse.options[:environment]
      end

      it 'sets concurrency' do
        assert_equal 50, Roundhouse.options[:concurrency]
      end

      it 'sets pid file' do
        assert_equal '/tmp/roundhouse-config-test.pid', Roundhouse.options[:pidfile]
      end

      it 'sets logfile' do
        assert_equal '/tmp/roundhouse.log', Roundhouse.options[:logfile]
      end
    end

    describe 'with env based config file' do
      before do
        @cli.parse(['roundhouse', '-e', 'staging', '-C', './test/env_based_config.yml'])
      end

      it 'takes a path' do
        assert_equal './test/env_based_config.yml', Roundhouse.options[:config_file]
      end

      it 'sets verbose' do
        refute Roundhouse.options[:verbose]
      end

      it 'sets require file' do
        assert_equal './test/fake_env.rb', Roundhouse.options[:require]
      end

      it 'sets environment' do
        assert_equal 'staging', Roundhouse.options[:environment]
      end

      it 'sets concurrency' do
        assert_equal 5, Roundhouse.options[:concurrency]
      end

      it 'sets pid file' do
        assert_equal '/tmp/roundhouse-config-test.pid', Roundhouse.options[:pidfile]
      end

      it 'sets logfile' do
        assert_equal '/tmp/roundhouse.log', Roundhouse.options[:logfile]
      end
    end

    describe 'with an empty config file' do
      before do
        @tmp_file = Tempfile.new('roundhouse-test')
        @tmp_path = @tmp_file.path
        @tmp_file.close!
      end

      after do
        File.unlink @tmp_path if File.exist? @tmp_path
      end

      it 'takes a path' do
        @cli.parse(['roundhouse', '-C', @tmp_path])
        assert_equal @tmp_path, Roundhouse.options[:config_file]
      end

      it 'should have an identical options hash, except for config_file' do
        @cli.parse(['roundhouse'])
        old_options = Roundhouse.options.clone

        @cli.parse(['roundhouse', '-C', @tmp_path])
        new_options = Roundhouse.options.clone
        refute_equal old_options, new_options

        new_options.delete(:config_file)
        assert_equal old_options, new_options
      end
    end

    describe 'with config file and flags' do
      before do
        # We need an actual file here.
        @tmp_lib_path = '/tmp/require-me.rb'
        File.open(@tmp_lib_path, 'w') do |f|
          f.puts "# do work"
        end

        @tmp_file = Tempfile.new('roundhouser')
        @tmp_path = @tmp_file.path
        @tmp_file.close!

        @cli.parse(['roundhouse',
                    '-C', './test/config.yml',
                    '-e', 'snoop',
                    '-c', '100',
                    '-r', @tmp_lib_path,
                    '-P', @tmp_path])
      end

      after do
        File.unlink @tmp_lib_path if File.exist? @tmp_lib_path
        File.unlink @tmp_path if File.exist? @tmp_path
      end

      it 'uses concurrency flag' do
        assert_equal 100, Roundhouse.options[:concurrency]
      end

      it 'uses require file flag' do
        assert_equal @tmp_lib_path, Roundhouse.options[:require]
      end

      it 'uses environment flag' do
        assert_equal 'snoop', Roundhouse.options[:environment]
      end

      it 'uses pidfile flag' do
        assert_equal @tmp_path, Roundhouse.options[:pidfile]
      end
    end
  end

  describe 'misc' do
    it 'handles interrupts' do
      cli = Roundhouse::CLI.new
      assert_raises Interrupt do
        cli.handle_signal('INT')
      end
      assert_raises Interrupt do
        cli.handle_signal('TERM')
      end
      cli.handle_signal('USR2')
      cli.handle_signal('TTIN')
    end

    it 'can fire events' do
      count = 0
      Roundhouse.options[:lifecycle_events][:startup] = [proc {
        count += 1
      }]
      cli = Roundhouse::CLI.new
      cli.fire_event(:startup)
      assert_equal 1, count
    end
  end

end
