require_relative 'helper'
require 'roundhouse/client'
require 'roundhouse/worker'
require 'roundhouse/api'

require 'active_support/core_ext/numeric/time'

class TestClient < Roundhouse::Test
  describe 'with mock redis' do
    before do
      @redis = Minitest::Mock.new
      def @redis.multi; [yield] * 2 if block_given?; end
      def @redis.set(*); true; end
      def @redis.sadd(*); true; end
      def @redis.srem(*); true; end
      def @redis.get(*); nil; end
      def @redis.del(*); nil; end
      def @redis.incrby(*); nil; end
      def @redis.setex(*); true; end
      def @redis.expire(*); true; end
      def @redis.watch(*); true; end
      def @redis.with_connection; yield self; end
      def @redis.with; yield self; end
      def @redis.exec; true; end
      def @redis.hset(*); true; end
      def @redis.hget(*); true; end
      Roundhouse.instance_variable_set(:@redis, @redis)
      Roundhouse::Client.instance_variable_set(:@default, nil)
    end

    after do
      Roundhouse.redis = REDIS
      Roundhouse::Client.instance_variable_set(:@default, nil)
    end

    it 'raises ArgumentError with invalid params' do
      assert_raises ArgumentError do
        Roundhouse::Client.push('foo', 1)
      end

      assert_raises ArgumentError do
        Roundhouse::Client.push('foo', :class => 'Foo', :noargs => [1, 2])
      end

      assert_raises ArgumentError do
        Roundhouse::Client.push('queue_id' => 'wrong', 'class' => MyWorker, 'noargs' => [1, 2])
      end

      assert_raises ArgumentError do
        Roundhouse::Client.push('queue_id' => 1, 'class' => MyWorker, 'noargs' => [1, 2])
      end

      assert_raises ArgumentError do
        Roundhouse::Client.push('queue_id' => 1, 'class' => 42, 'args' => [1, 2])
      end

      assert_raises ArgumentError do
        Roundhouse::Client.push('queue_id' => 1, 'class' => MyWorker, 'args' => 1)
      end
    end

    describe 'as instance' do
      it 'can push' do
        @redis.expect :lpush, 1, ['queue:1', Array]
        client = Roundhouse::Client.new
        Roundhouse::Monitor.stub :maybe_add_to_rotation, true do
          jid = client.push('class' => 'Blah', 'args' => [1,2,3], 'queue_id' => 1)
          assert_equal 24, jid.size
        end
      end

      it 'allows local middleware modification' do
        @redis.expect :lpush, 1, ['queue:1', Array]
        $called = false
        mware = Class.new { def call(worker_klass,msg,q,r); $called = true; msg;end }
        client = Roundhouse::Client.new
        client.middleware do |chain|
          chain.add mware
        end

        Roundhouse::Monitor.stub :maybe_add_to_rotation, true do
          client.push('class' => 'Blah', 'args' => [1,2,3], 'queue_id' => 1)
        end

        assert $called
        assert client.middleware.exists?(mware)
        refute Roundhouse.client_middleware.exists?(mware)
      end
    end

    it 'pushes messages to redis' do
      @redis.expect :lpush, 1, ['queue:1', Array]
      Roundhouse::Monitor.stub :maybe_add_to_rotation, true do
        pushed = Roundhouse::Client.push('queue_id' => 1, 'class' => MyWorker, 'args' => [1, 2])
        assert pushed
        assert_equal 24, pushed.size
      end
      @redis.verify
    end

    it 'pushes messages to redis using a String class' do
      @redis.expect :lpush, 1, ['queue:1', Array]
      Roundhouse::Monitor.stub :maybe_add_to_rotation, true do
        pushed = Roundhouse::Client.push('queue_id' => 1, 'class' => 'MyWorker', 'args' => [1, 2])
        assert pushed
        assert_equal 24, pushed.size
      end
      @redis.verify
    end

    class MyWorker
      include Roundhouse::Worker
    end

    it 'has default options' do
      assert_equal Roundhouse.default_worker_options, MyWorker.get_roundhouse_options
    end

    it 'handles perform_async' do
      @redis.expect :lpush, 1, ['queue:100', Array]
      Roundhouse::Monitor.stub :maybe_add_to_rotation, true do
        pushed = MyWorker.perform_async(100, 1, 2)
        assert pushed
      end
      @redis.verify
    end

    it 'enqueues messages to redis' do
      @redis.expect :lpush, 1, ['queue:100', Array]
      Roundhouse::Monitor.stub :maybe_add_to_rotation, true do
        pushed = Roundhouse::Client.enqueue_to(100, MyWorker, 1, 2)
        assert pushed
      end
      @redis.verify
    end

    it 'enqueues messages to redis (delayed)' do
      @redis.expect :zadd, 1, ['schedule', Array]
      Roundhouse::Monitor.stub :maybe_add_to_rotation, true do
        pushed = Roundhouse::Client.enqueue_to_in(100, 3.minutes, MyWorker, 1, 2)
        assert pushed
      end
      @redis.verify
    end

    it 'enqueues messages to redis (delayed into past)' do
      @redis.expect :lpush, 1, ['queue:100', Array]
      Roundhouse::Monitor.stub :maybe_add_to_rotation, true do
        pushed = Roundhouse::Client.enqueue_to_in(100, -3.minutes, MyWorker, 1, 2)
        assert pushed
      end
      @redis.verify
    end

    class QueuedWorker
      include Roundhouse::Worker
      roundhouse_options :queue => :flimflam
    end

    it 'enqueues to the named queue' do
      @redis.expect :lpush, 1, ['queue:100', Array]
      Roundhouse::Monitor.stub :maybe_add_to_rotation, true do
        pushed = QueuedWorker.perform_async(100, 1, 2)
        assert pushed
      end
      @redis.verify
    end

    it 'retrieves queues' do
      @redis.expect :smembers, ['0'], [Roundhouse::Monitor::BUCKETS]
      @redis.expect :hkeys, ['100'], [Roundhouse::Monitor.status_bucket(100)]
      assert_equal [100], Roundhouse::Queue.all.map(&:queue_id)
    end
  end

  describe 'bulk' do
    after do
      Roundhouse::Queue.new(100).clear
    end
    it 'can push a large set of jobs at once' do
      jids = Roundhouse::Client.push_bulk('class' => QueuedWorker, 'queue_id' => 100, 'args' => (1..1_000).to_a.map { |x| Array(x) })
      assert_equal 1_000, jids.size
    end

    it 'can push a large set of jobs at once using a String class' do
      jids = Roundhouse::Client.push_bulk('class' => 'QueuedWorker', 'queue_id' => 100, 'args' => (1..1_000).to_a.map { |x| Array(x) })
      assert_equal 1_000, jids.size
    end
    it 'returns the jids for the jobs' do
      Roundhouse::Client.push_bulk('class' => 'QueuedWorker', 'queue_id' => 100, 'args' => (1..2).to_a.map { |x| Array(x) }).each do |jid|
        assert_match(/[0-9a-f]{12}/, jid)
      end
    end
  end

  class BaseWorker
    include Roundhouse::Worker
    roundhouse_options 'retry' => 'base'
  end
  class AWorker < BaseWorker
  end
  class BWorker < BaseWorker
    roundhouse_options 'retry' => 'b'
  end
  class CWorker < BaseWorker
    roundhouse_options 'retry' => 2
  end

  describe 'client middleware' do

    class Stopper
      def call(worker_class, message, queue, r)
        raise ArgumentError unless r
        yield if message['args'].first.odd?
      end
    end

    it 'can stop some of the jobs from pushing' do
      Roundhouse.client_middleware.add Stopper
      begin
        assert_equal nil, Roundhouse::Client.push('queue_id' => 100, 'class' => MyWorker, 'args' => [0])
        assert_match(/[0-9a-f]{12}/, Roundhouse::Client.push('queue_id' => 100, 'class' => MyWorker, 'args' => [1]))
        Roundhouse::Client.push_bulk('queue_id' => 100, 'class' => MyWorker, 'args' => [[0], [1]]).each do |jid|
          assert_match(/[0-9a-f]{12}/, jid)
        end
      ensure
        Roundhouse.client_middleware.remove Stopper
      end
    end
  end

  describe 'inheritance' do
    it 'inherits roundhouse options' do
      assert_equal 'base', AWorker.get_roundhouse_options['retry']
      assert_equal 'b', BWorker.get_roundhouse_options['retry']
    end
  end

  describe 'item normalization' do
    it 'defaults retry to true' do
      assert_equal true, Roundhouse::Client.new.__send__(:normalize_item, 'queue_id' => 100, 'class' => QueuedWorker, 'args' => [])['retry']
    end

    it "does not normalize numeric retry's" do
      assert_equal 2, Roundhouse::Client.new.__send__(:normalize_item, 'queue_id' => 100, 'class' => CWorker, 'args' => [])['retry']
    end
  end

  describe 'sharding' do
    class DWorker < BaseWorker
    end

    it 'allows roundhouse_options to point to different Redi' do
      skip 'sharding not needed right now'

      conn = MiniTest::Mock.new
      conn.expect(:multi, [0, 1])
      DWorker.roundhouse_options('pool' => ConnectionPool.new(size: 1) { conn })
      DWorker.perform_async(1,2,3)
      conn.verify
    end

    it 'allows #via to point to different Redi' do
      skip '#via not needed right now'

      conn = MiniTest::Mock.new
      conn.expect(:multi, [0, 1])
      default = Roundhouse::Client.new.redis_pool
      sharded_pool = ConnectionPool.new(size: 1) { conn }
      Roundhouse::Client.via(sharded_pool) do
        CWorker.perform_async(100, 1,2,3)
        assert_equal sharded_pool, Roundhouse::Client.new.redis_pool
        assert_raises RuntimeError do
          Roundhouse::Client.via(default) do
            # nothing
          end
        end
      end
      assert_equal default, Roundhouse::Client.new.redis_pool
      conn.verify
    end

    it 'allows Resque helpers to point to different Redi' do
      skip 'sharding not needed right now'

      conn = MiniTest::Mock.new
      conn.expect(:multi, []) { |*args, &block| block.call }
      conn.expect(:zadd, 1, [String, Array])
      DWorker.roundhouse_options('pool' => ConnectionPool.new(size: 1) { conn })
      Roundhouse::Client.enqueue_to_in(100, 10, DWorker, 3)
      conn.verify
    end
  end
end
