require_relative 'helper'
require 'roundhouse/fetch'
require 'roundhouse/monitor'
require 'roundhouse/api'

class TestFetcher < Roundhouse::Test
  describe 'fetcher' do
    before do
      Roundhouse.redis = { :url => REDIS_URL, :namespace => 'fuzzy' }
      Roundhouse.redis { |conn| conn.flushdb }
    end

    after do
      Roundhouse.redis = REDIS
    end

    describe '#retrieve_work' do
      describe 'when nominal' do
        before do
          Roundhouse.redis do |conn|
            conn.lpush(Roundhouse::Monitor::TURNTABLE, 100)
            conn.hset(Roundhouse::Monitor.status_bucket(100), 100, Roundhouse::Monitor::ACTIVE)
            conn.lpush("#{Roundhouse::Monitor::QUEUE}:100", 'msg')
          end
        end

        it 'retrieves' do
          fetch = Roundhouse::RoundRobinFetch.new
          uow = fetch.retrieve_work
          refute_nil uow
          assert_equal '100', uow.queue_id
          assert_equal 'msg', uow.message
          q = Roundhouse::Queue.new(100)
          assert_equal 0, q.size
          uow.requeue
          assert_equal 1, q.size
          assert_nil uow.acknowledge
        end
      end

      describe 'with empty status' do
        before do
          Roundhouse.redis do |conn|
            conn.lpush(Roundhouse::Monitor::TURNTABLE, 100)
            conn.lpush(Roundhouse::Monitor::TURNTABLE, 200)
            conn.hset(Roundhouse::Monitor.status_bucket(100), 100, Roundhouse::Monitor::EMPTY)
            conn.hset(Roundhouse::Monitor.status_bucket(200), 200, Roundhouse::Monitor::ACTIVE)
            conn.lpush("#{Roundhouse::Monitor::QUEUE}:200", 'msg-200')
          end
        end

        it 'should return nil' do
          fetch = Roundhouse::RoundRobinFetch.new
          uow = fetch.retrieve_work
          assert_nil uow
        end
      end

      describe 'with active status and empty queue' do
        before do
          Roundhouse.redis do |conn|
            conn.lpush(Roundhouse::Monitor::TURNTABLE, 100)
            conn.lpush(Roundhouse::Monitor::TURNTABLE, 200)
            conn.hset(Roundhouse::Monitor.status_bucket(100), 100, Roundhouse::Monitor::ACTIVE)
            conn.hset(Roundhouse::Monitor.status_bucket(200), 200, Roundhouse::Monitor::ACTIVE)
            conn.lpush("#{Roundhouse::Monitor::QUEUE}:200", 'msg-200')
          end
        end

        it 'should return nil' do
          empty_queue = Roundhouse::Queue.new(100)
          assert_equal 0, empty_queue.size
          assert_equal :active, empty_queue.status

          fetch = Roundhouse::RoundRobinFetch.new
          uow = fetch.retrieve_work
          assert_nil uow
        end
      end

      describe 'with suspended status' do
        before do
          Roundhouse.redis do |conn|
            conn.lpush(Roundhouse::Monitor::TURNTABLE, 100)
            conn.lpush(Roundhouse::Monitor::TURNTABLE, 200)
            conn.hset(Roundhouse::Monitor.status_bucket(100), 100, Roundhouse::Monitor::SUSPENDED)
            conn.hset(Roundhouse::Monitor.status_bucket(200), 200, Roundhouse::Monitor::ACTIVE)
            conn.lpush("#{Roundhouse::Monitor::QUEUE}:100", 'msg-100')
            conn.lpush("#{Roundhouse::Monitor::QUEUE}:200", 'msg-200')
          end
        end

        it 'should return nil' do
          fetch = Roundhouse::RoundRobinFetch.new
          uow = fetch.retrieve_work
          assert_nil uow
        end
      end

    end

    it 'bulk requeues' do
      q1 = Roundhouse::Queue.new(1000)
      q2 = Roundhouse::Queue.new(2000)
      assert_equal 0, q1.size
      assert_equal 0, q2.size
      uow = Roundhouse::RoundRobinFetch::UnitOfWork
      Roundhouse::RoundRobinFetch.bulk_requeue([uow.new('fuzzy:queue:1000', 'bob'), uow.new('fuzzy:queue:1000', 'bar'), uow.new('fuzzy:queue:2000', 'widget')], {:queues => []})
      assert_equal 2, q1.size
      assert_equal 1, q2.size

      # Changed bulk requeues not to change status
      # Not sure if this is a good idea yet
      #assert_equal :active, q1.status
      #assert_equal :active, q2.status
    end

    describe 'fetching' do
      before do
        Roundhouse::Fetcher.reset
      end

      it 'instantiates' do
        begin
          Roundhouse.options[:fetch] = NullFetch
          mgr = Minitest::Mock.new
          fetch = Roundhouse::Fetcher.new(mgr, {})
          fetch.fetch
          Roundhouse::Fetcher.done!
        ensure
          Roundhouse.options[:fetch] = Roundhouse::RoundRobinFetch
        end
      end

      class NullFetch
        def initialize(opts)
        end
        def retrieve_work
        end
        def self.bulk_requeue(*args)
        end
      end

      it 'handles redis network errors' do
        begin
          Roundhouse.logger.level = Logger::FATAL
          Roundhouse.options[:fetch] = ErrorFetch
          mgr = Minitest::Mock.new
          fetch = Roundhouse::Fetcher.new(mgr, {})
          def fetch.pause
          end
          refute fetch.down
          fetch.fetch
          Roundhouse::Fetcher.done!
          assert fetch.down
        ensure
          Roundhouse.options[:fetch] = Roundhouse::RoundRobinFetch
          Roundhouse.logger.level = Logger::ERROR
        end
      end

      class ErrorFetch
        def initialize(opts)
        end
        def retrieve_work
          raise IOError, "ker-BOOM"
        end
        def self.bulk_requeue(*args)
        end
      end
    end

  end
end
