require_relative 'helper'
#require 'roundhouse/api'
require 'roundhouse/monitor'

require 'active_support/core_ext/numeric/time'

class TestMonitor < Roundhouse::Test
  describe 'Roundhouse::Monitor' do
    before do
      Roundhouse.redis(&:flushdb)
    end

    describe '#pop' do
      describe 'when nominal' do
        it 'should pop the next available job from the next available queue' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.push_job conn, [{'queue_id' => 100, 'class' => 'MyWorker', 'args' => [1]}]
            assert conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)

            assert_equal '100', Roundhouse::Monitor.pop(conn)
            refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)
          end
        end
      end

      describe 'with queue with empty status' do
        it 'should return nil' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.activate(conn, '200')
            Roundhouse::Monitor.push(conn, '200')
            Roundhouse::Monitor.set_queue_is_empty(conn, '200')
            Roundhouse::Monitor.push_job conn, [{'queue_id' => 100, 'class' => 'MyWorker', 'args' => [1]}]
            assert conn.sismember(Roundhouse::Monitor::IN_ROTATION, 200)

            assert_equal nil, Roundhouse::Monitor.pop(conn)
            refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, 200)
          end
        end
      end

      describe 'with empty queue with active status' do
        it 'should return queue_id' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.activate(conn, '200')
            Roundhouse::Monitor.push(conn, '200')
            Roundhouse::Monitor.push_job conn, [{'queue_id' => 100, 'class' => 'MyWorker', 'args' => [1]}]
            assert conn.sismember(Roundhouse::Monitor::IN_ROTATION, 200)

            assert_equal '200', Roundhouse::Monitor.pop(conn)
            refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, 200)
          end
        end
      end

      describe 'with queue with suspended status' do
        it 'should return nil' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.activate(conn, '200')
            Roundhouse::Monitor.push(conn, '200')
            Roundhouse::Monitor.suspend(conn, '200')
            Roundhouse::Monitor.push_job conn, [{'queue_id' => 100, 'class' => 'MyWorker', 'args' => [1]}]
            assert conn.sismember(Roundhouse::Monitor::IN_ROTATION, 200)

            assert_equal nil, Roundhouse::Monitor.pop(conn)
            refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, 200)
          end
        end
      end

      describe 'when nothing is in the turntable' do
        it "should block for #{Roundhouse::Monitor::TURNTABLE_TIMEOUT} seconds and return nil" do
          skip 'skipping slow tests' unless ENV['SLOW']
          Roundhouse.redis do |conn|
            assert_equal nil, Roundhouse::Monitor.pop(conn)
          end
        end
      end
    end # #pop

    describe '#push' do
      describe 'with active queue' do
        it 'should push into turntable' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.push_job conn, [{'queue_id' => 100, 'class' => 'MyWorker', 'args' => [1]}]

            assert_equal ['100'], conn.lrange(Roundhouse::Monitor::TURNTABLE, -1, -1)
            assert conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)
          end
        end

        describe 'when already in turntable' do
          it 'should not push into turntable' do
            Roundhouse.redis do |conn|
              Roundhouse::Monitor.activate(conn, 100)
              Roundhouse::Monitor.push(conn, 100)
              assert_equal ['100'], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)
              assert conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)

              Roundhouse::Monitor.push_job conn, [{'queue_id' => 100, 'class' => 'MyWorker', 'args' => [1]}]
              assert_equal ['100'], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)
              assert conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)
            end
          end
        end
      end

      describe 'with queue with empty status' do
        it 'should not push into turntable' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.set_queue_is_empty(conn, '200')

            assert_equal [], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)
            refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)
          end
        end
      end

      describe 'with queue with suspended status' do
        it 'should skip' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.suspend(conn, '200')
            Roundhouse::Monitor.push(conn, '200')

            assert_equal [], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)
            refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)
          end
        end
      end
    end # #push

    describe '#pop_job' do
      describe 'with jobs in queue' do
        it 'should pop the next job and return data' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.push_job conn, [{'queue_id' => 100, 'class' => 'MyWorker', 'args' => [1]}]
            assert conn.lrange("#{Roundhouse::Monitor::QUEUE}:100", 0, 0).first

            assert Roundhouse::Monitor.pop_job conn, 100

            # Should still be active
            assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
          end
        end
      end

      describe 'without jobs in queue' do
        it 'should return nil' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.activate(conn, 100)
            assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
            refute conn.lrange("#{Roundhouse::Monitor::QUEUE}:100", 0, 0).first

            refute Roundhouse::Monitor.pop_job conn, 100

            # Should be set to inactive
            assert_equal Roundhouse::Monitor::EMPTY, Roundhouse::Monitor.queue_status(conn, 100)
          end
        end
      end
    end

    describe '#push_job' do
      it 'should add bucket and push to queue' do
        Roundhouse.redis do |conn|
          bucket_num = Roundhouse::Monitor.bucket_num(2500)
          refute conn.sismember(Roundhouse::Monitor::BUCKETS, bucket_num)
          refute conn.lrange("#{Roundhouse::Monitor::QUEUE}:2500", 0, 0).first

          Roundhouse::Monitor.push_job conn, [{'queue_id' => 2500, 'class' => 'MyWorker', 'args' => [1]}]

          # We should see the bucket registered
          assert conn.sismember(Roundhouse::Monitor::BUCKETS, bucket_num)
          assert conn.lrange("#{Roundhouse::Monitor::QUEUE}:2500", 0, 0).first
        end
      end

    end

    describe '#requeue' do
      it 'should requeue items'
    end

    describe '#maybe_add_to_rotation' do
      describe 'with active queue' do
        it 'should not rotate queue into turntable' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.activate(conn, 100)
            assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
            assert_equal [], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, 0)
            refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)

            Roundhouse::Monitor.maybe_add_to_rotation(conn, 100)
            assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
            assert_equal [], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, 0)
            refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)
          end
        end
      end

      describe 'with empty queue' do
        it 'should rotate queue into turntable' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.set_queue_is_empty(conn, 100)
            assert_equal Roundhouse::Monitor::EMPTY, Roundhouse::Monitor.queue_status(conn, 100)
            assert_equal [], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)
            refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)

            Roundhouse::Monitor.maybe_add_to_rotation(conn, 100)
            assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
            assert_equal ['100'], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, 0)
            assert conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)
          end
        end

        describe 'when already in rotation' do
          it 'should not rotate queue into turntable' do
            Roundhouse.redis do |conn|
              Roundhouse::Monitor.activate(conn, 100)
              Roundhouse::Monitor.push(conn, 100)
              Roundhouse::Monitor.set_queue_is_empty(conn, 100)

              assert_equal Roundhouse::Monitor::EMPTY, Roundhouse::Monitor.queue_status(conn, 100)
              assert_equal ['100'], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)
              assert conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)

              Roundhouse::Monitor.maybe_add_to_rotation(conn, 100)
              assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
              assert_equal ['100'], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)
              assert conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)
            end
          end

        end
      end

      describe 'with suspended queue' do
        it 'should not rotate queue into semaphore' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.suspend(conn, 100)
            assert_equal Roundhouse::Monitor::SUSPENDED, Roundhouse::Monitor.queue_status(conn, 100)
            assert_equal [], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)
            refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)

            Roundhouse::Monitor.maybe_add_to_rotation(conn, 100)
            assert_equal Roundhouse::Monitor::SUSPENDED, Roundhouse::Monitor.queue_status(conn, 100)
            assert_equal [], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)
            refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, 100)
          end
        end
      end
    end # maybe_add_to_rotation

    describe '#set_queue_is_empty' do
      it 'should set queue as empty' do
        Roundhouse.redis do |conn|
          Roundhouse::Monitor.set_queue_is_empty(conn, 100)
          assert_equal Roundhouse::Monitor::EMPTY, Roundhouse::Monitor.queue_status(conn, 100)
        end
      end
    end # #set_queue_is_empty

    describe '#activate' do
      it 'should set queue as active' do
        Roundhouse.redis do |conn|
          Roundhouse::Monitor.activate(conn, 100)
          assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
        end
      end
    end # #activate

    describe '#suspend' do
      it 'should set queue as suspended' do
        Roundhouse.redis do |conn|
          Roundhouse::Monitor.suspend(conn, 100)
          assert_equal Roundhouse::Monitor::SUSPENDED, Roundhouse::Monitor.queue_status(conn, 100)
        end
      end
    end # #suspend

    describe '#resume' do
      describe 'with suspended queue' do
        it 'should reactivate queue and put it into rotation' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.suspend(conn, 100)
            assert_equal Roundhouse::Monitor::SUSPENDED, Roundhouse::Monitor.queue_status(conn, 100)
            assert_equal [], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)

            Roundhouse::Monitor.resume(conn, 100)
            assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
            assert_equal ['100'], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)
          end
        end

        describe 'when queue is already in rotation' do
          it 'should reactivate but not put into queue' do
            Roundhouse.redis do |conn|
              Roundhouse::Monitor.activate(conn, 100)
              Roundhouse::Monitor.push(conn, 100) # Puts it into rotation

              Roundhouse::Monitor.suspend(conn, 100)
              assert_equal Roundhouse::Monitor::SUSPENDED, Roundhouse::Monitor.queue_status(conn, 100)
              assert_equal ['100'], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)

              Roundhouse::Monitor.resume(conn, 100)
              assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
              # There should only be a single queue 100 in the turntable
              assert_equal ['100'], conn.lrange(Roundhouse::Monitor::TURNTABLE, 0, -1)
            end
          end
        end
      end

      describe 'with empty queue' do
        it 'should not resume' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.set_queue_is_empty(conn, 100)
            assert_equal Roundhouse::Monitor::EMPTY, Roundhouse::Monitor.queue_status(conn, 100)
            assert_equal [], conn.lrange(Roundhouse::Monitor::TURNTABLE, -1, -1)

            Roundhouse::Monitor.resume(conn, 100)
            assert_equal Roundhouse::Monitor::EMPTY, Roundhouse::Monitor.queue_status(conn, 100)
            assert_equal [], conn.lrange(Roundhouse::Monitor::TURNTABLE, -1, -1)
          end
        end
      end

      describe 'with active queue' do
        it 'should not resume' do
          Roundhouse.redis do |conn|
            Roundhouse::Monitor.activate(conn, 100)
            assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
            assert_equal [], conn.lrange(Roundhouse::Monitor::TURNTABLE, -1, -1)

            Roundhouse::Monitor.resume(conn, 100)
            assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
            assert_equal [], conn.lrange(Roundhouse::Monitor::TURNTABLE, -1, -1)
          end
        end
      end
    end # #resume

    describe '#rebuild_turntable!' do
      describe 'with active-status queue' do
        before do
          Roundhouse.redis { |conn| Roundhouse::Monitor.activate(conn, 100) }
        end

        describe 'with someting in the queue' do
          it 'should put queue into turntable' do
            Roundhouse.redis { |conn| conn.lpush("#{Roundhouse::Monitor::QUEUE}:100", 'something') }
            Roundhouse::Monitor.rebuild_turntable!

            Roundhouse.redis do |conn|
              # Should still be active
              assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
              # Should be in turntable
              assert_equal 1, conn.llen(Roundhouse::Monitor::TURNTABLE)
              assert conn.sismember(Roundhouse::Monitor::IN_ROTATION, '100')
              assert_equal '100', conn.rpop(Roundhouse::Monitor::TURNTABLE)
            end
          end

        end

        describe 'without anything in queue' do
          it 'should set queue to empty and not put into turntable' do
            Roundhouse::Monitor.rebuild_turntable!

            Roundhouse.redis do |conn|
              # Should set to empty
              assert_equal Roundhouse::Monitor::EMPTY, Roundhouse::Monitor.queue_status(conn, 100)
              # Should not be in turntable
              assert_equal 0, conn.llen(Roundhouse::Monitor::TURNTABLE)
              refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, '100')
            end
          end
        end
      end

      describe 'with empty-status queue' do
        before do
          Roundhouse.redis { |conn| Roundhouse::Monitor.set_queue_is_empty(conn, 100) }
        end

        describe 'with someting in the queue' do
          it 'should activate queue and put queue into turntable' do
            Roundhouse.redis { |conn| conn.lpush("#{Roundhouse::Monitor::QUEUE}:100", 'something') }
            Roundhouse::Monitor.rebuild_turntable!

            Roundhouse.redis do |conn|
              # Should be set to active
              assert_equal Roundhouse::Monitor::ACTIVE, Roundhouse::Monitor.queue_status(conn, 100)
              # Should be in turntable
              assert_equal 1, conn.llen(Roundhouse::Monitor::TURNTABLE)
              assert conn.sismember(Roundhouse::Monitor::IN_ROTATION, '100')
              assert_equal '100', conn.rpop(Roundhouse::Monitor::TURNTABLE)
            end
          end

        end

        describe 'without anything in queue' do
          it 'should not put into turntable' do
            Roundhouse::Monitor.rebuild_turntable!

            Roundhouse.redis do |conn|
              # Should still be empty
              assert_equal Roundhouse::Monitor::EMPTY, Roundhouse::Monitor.queue_status(conn, 100)
              # Should not be in turntable
              assert_equal 0, conn.llen(Roundhouse::Monitor::TURNTABLE)
              refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, '100')
            end
          end
        end
      end

      describe 'with suspended-status queue' do
        before do
          Roundhouse.redis { |conn| Roundhouse::Monitor.suspend(conn, 100) }
        end

        describe 'with something in the queue' do
          it 'should not put queue into turntable' do
            Roundhouse.redis { |conn| conn.lpush("#{Roundhouse::Monitor::QUEUE}:100", 'something') }
            Roundhouse::Monitor.rebuild_turntable!

            Roundhouse.redis do |conn|
              # Should still be suspended
              assert_equal Roundhouse::Monitor::SUSPENDED, Roundhouse::Monitor.queue_status(conn, 100)
              # Should not be in turntable
              assert_equal 0, conn.llen(Roundhouse::Monitor::TURNTABLE)
              refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, '100')
            end
          end

        end

        describe 'without anything in queue' do
          it 'should not put into turntable' do
            Roundhouse::Monitor.rebuild_turntable!

            Roundhouse.redis do |conn|
              # Should still be suspended
              assert_equal Roundhouse::Monitor::SUSPENDED, Roundhouse::Monitor.queue_status(conn, 100)
              # Should not be in turntable
              assert_equal 0, conn.llen(Roundhouse::Monitor::TURNTABLE)
              refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, '100')
            end
          end
        end
      end

      describe 'with unknown-status queue' do
        before do
          Roundhouse.redis { |conn| Roundhouse::Monitor.send(:set_queue_status, conn, 100, -2000) }
        end

        describe 'with something in the queue' do
          it 'should not put queue into turntable' do
            Roundhouse.redis { |conn| conn.lpush("#{Roundhouse::Monitor::QUEUE}:100", 'something') }
            Roundhouse::Monitor.rebuild_turntable!

            Roundhouse.redis do |conn|
              # Should be suspended
              assert_equal Roundhouse::Monitor::SUSPENDED, Roundhouse::Monitor.queue_status(conn, 100)
              # Should not be in turntable
              assert_equal 0, conn.llen(Roundhouse::Monitor::TURNTABLE)
              refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, '100')
            end
          end

        end

        describe 'without anything in queue' do
          it 'should not put into turntable' do
            Roundhouse::Monitor.rebuild_turntable!

            Roundhouse.redis do |conn|
              # Should be suspended
              assert_equal Roundhouse::Monitor::SUSPENDED, Roundhouse::Monitor.queue_status(conn, 100)
              # Should not be in turntable
              assert_equal 0, conn.llen(Roundhouse::Monitor::TURNTABLE)
              refute conn.sismember(Roundhouse::Monitor::IN_ROTATION, '100')
            end
          end
        end
      end

      it 'scaling test' do
        skip 'slow test' unless ENV['SLOW']
        require 'benchmark'

        num_queues = 10_000
        num_active = (num_queues * 0.7).to_i
        num_empty  = (num_queues * 0.2).to_i
        num_suspended = num_queues - num_active - num_empty
        active_with_empty = (num_active * 0.25).to_i
        active_with_active = num_active - active_with_empty
        empty_with_something = (num_empty * 0.25).to_i
        empty_with_empty = num_empty - empty_with_something

        states = \
          Array.new(active_with_empty)    { :active_with_empty } +
          Array.new(active_with_active)   { :active_with_active } +
          Array.new(empty_with_something) { :empty_with_something } +
          Array.new(empty_with_empty)     { :empty_with_empty } +
          Array.new(num_suspended)        { :suspended }

        rand_msgs = -> { Array.new(rand(3) + 1) { 'msg' } }
        puts "\nScaling Test:\nSetting up"
        Roundhouse.redis do |conn|
          conn.multi do
            states.shuffle.each_with_index do |state, q_id|
              case state
              when :active_with_empty then
                Roundhouse::Monitor.activate(conn, q_id)
              when :empty_with_empty then
                Roundhouse::Monitor.set_queue_is_empty(conn, q_id)
              when :suspended then
                Roundhouse::Monitor.set_queue_is_empty(conn, q_id)
              when :active_with_active then
                Roundhouse::Monitor.activate(conn, q_id)
                conn.lpush("#{Roundhouse::Monitor::QUEUE}:#{q_id}", rand_msgs.())
              when :empty_with_something then
                Roundhouse::Monitor.set_queue_is_empty(conn, q_id)
                conn.lpush("#{Roundhouse::Monitor::QUEUE}:#{q_id}", rand_msgs.())
              end
            end
          end
        end

        res = Benchmark.measure { Roundhouse::Monitor.rebuild_turntable! }
        puts res, ''
      end
    end # #rebuild_turntable!
  end
end
