require_relative 'helper'
require 'roundhouse/scheduled'

class TestScheduled < Roundhouse::Test
  class ScheduledWorker
    include Roundhouse::Worker
    def perform(x)
    end
  end

  describe 'poller' do
    before do
      Roundhouse.redis = REDIS
      Roundhouse.redis do |conn|
        conn.flushdb
      end

      @error_1  = { 'class' => ScheduledWorker.name, 'args' => [0], 'queue_id' => 100 }
      @error_2  = { 'class' => ScheduledWorker.name, 'args' => [1], 'queue_id' => 200 }
      @error_3  = { 'class' => ScheduledWorker.name, 'args' => [2], 'queue_id' => 300 }
      @future_1 = { 'class' => ScheduledWorker.name, 'args' => [3], 'queue_id' => 400 }
      @future_2 = { 'class' => ScheduledWorker.name, 'args' => [4], 'queue_id' => 500 }
      @future_3 = { 'class' => ScheduledWorker.name, 'args' => [5], 'queue_id' => 600 }

      @retry = Roundhouse::RetrySet.new
      @scheduled = Roundhouse::ScheduledSet.new
      @poller = Roundhouse::Scheduled::Poller.new
    end

    class Stopper
      def call(worker_class, message, queue, r)
        yield if message['args'].first.odd?
      end
    end

    it 'executes client middleware' do
      Roundhouse.client_middleware.add Stopper
      begin
        @retry.schedule (Time.now - 60).to_f, @error_1
        @retry.schedule (Time.now - 60).to_f, @error_2
        @scheduled.schedule (Time.now - 60).to_f, @future_2
        @scheduled.schedule (Time.now - 60).to_f, @future_3

        @poller.poll

        Roundhouse.redis do |conn|
          assert_equal 0, conn.llen("queue:100")
          assert_equal 1, conn.llen("queue:200")
          assert_equal 0, conn.llen("queue:500")
          assert_equal 1, conn.llen("queue:600")
        end
      ensure
        Roundhouse.client_middleware.remove Stopper
      end
    end

    it 'should empty the retry and scheduled queues up to the current time' do
      created_time  = Time.new(2013, 2, 3)
      enqueued_time = Time.new(2013, 2, 4)

      Time.stub(:now, created_time) do
        @retry.schedule (enqueued_time - 60).to_f, @error_1.merge!('created_at' => created_time.to_f)
        @retry.schedule (enqueued_time - 50).to_f, @error_2.merge!('created_at' => created_time.to_f)
        @retry.schedule (enqueued_time + 60).to_f, @error_3.merge!('created_at' => created_time.to_f)
        @scheduled.schedule (enqueued_time - 60).to_f, @future_1.merge!('created_at' => created_time.to_f)
        @scheduled.schedule (enqueued_time - 50).to_f, @future_2.merge!('created_at' => created_time.to_f)
        @scheduled.schedule (enqueued_time + 60).to_f, @future_3.merge!('created_at' => created_time.to_f)
      end

      Time.stub(:now, enqueued_time) do
        @poller.poll

        Roundhouse.redis do |conn|
          %w(queue:100 queue:200 queue:400 queue:500).each do |queue_id|
            assert_equal 1, conn.llen(queue_id)
            job = Roundhouse.load_json(conn.lrange(queue_id, 0, -1)[0])
            assert_equal enqueued_time.to_f, job['enqueued_at']
            assert_equal created_time.to_f,  job['created_at']
          end
        end

        assert_equal 1, @retry.size
        assert_equal 1, @scheduled.size
      end
    end

    def with_roundhouse_option(name, value)
      _original, Roundhouse.options[name] = Roundhouse.options[name], value
      begin
        yield
      ensure
        Roundhouse.options[name] = _original
      end
    end

    it 'generates random intervals that target a configured average' do
      with_roundhouse_option(:poll_interval_average, 10) do
        i = 500
        intervals = i.times.map{ @poller.send(:random_poll_interval) }

        assert intervals.all?{|x| x >= 5}
        assert intervals.all?{|x| x <= 15}
        assert_in_delta 10, intervals.reduce(&:+).to_f / i, 0.5
      end
    end

    it 'calculates an average poll interval based on the number of known Roundhouse processes' do
      with_roundhouse_option(:average_scheduled_poll_interval, 10) do
        3.times do |i|
          Roundhouse.redis do |conn|
            conn.sadd("processes", "process-#{i}")
            conn.hset("process-#{i}", "info", nil)
          end
        end

        assert_equal 30, @poller.send(:scaled_poll_interval)
      end
    end
  end
end
